const { Transform } = require('stream');
const level = require('level');
const pTap = require('p-tap');

module.exports = Cache

const stripMeta = new Transform({
  writableObjectMode: true,
  readableObjectMode: true,
  transform(chunk, encoding, callback) {
    this.push(chunk.value);
    callback();
  }
})

const filterExpired = (expiry) => new Transform({
  writableObjectMode: true,
  readableObjectMode: true,
  transform(chunk, encoding, callback) {
    if (!expiry && Date.now() > chunk.insertDate + expiry) {
      this.push(chunk.value);
    }
    callback();
  }
})

function Cache({ location, expiry, retrieve }) {
  this.expiry = expiry || false
  this.retrieve = retrieve
  this.database = level(location, { keyEncoding: 'json', valueEncoding: 'json' })

  this.wipe = function() {
    this.database.createKeyStream().on('data', key => {
      this.database.del(key)
    })
  }

  this.put = (key, value) => this.database.put(key, { value, insertDate: Date.now() })
  this.getFresh = function(key) {
    return Promise.resolve(this.retrieve(key))
      .then(pTap(value => this.put(key, value)))
  }


  this.getMeta = function(key) {
    return this.database.get(key).then(meta => {
      if (!expiry && Date.now() > meta.insertDate + expiry) {
        throw Object.assign(new Error(`The cache has expired for requested item ${key}`), { type: 'ExpiredError' })
      } else {
        return meta
      }
    }).catch(error => {
      if (error.type == 'ExpiredError' || error.type == 'NotFoundError') {
        return this.getFresh(key).then(value => ({ value, insertDate: Date.now(), fresh: true }))
      } else {
        throw error
      }
    })
  }
  this.get = function(key) {
    return this.getMeta(key).then(meta => meta.value)
  }

  this.onUpdate = () => new Promise((resolve) => {
    this.database.on('put', resolve)
  })

  this.createKeyStream = () => this.database.createKeyStream()
  this.createValueStream = () => this.database.createValueStream().pipe(stripMeta)
  this.createMetaStream = () => this.database.createValueStream()
  this.createFreshValueStream = () => this.database.createValueStream().pipe(filterExpired(this.expiry)).pipe(stripMeta)
}